(* 
   StatVerif translator.

   Translates a process described in an extension (to handle explicit
   state) of ProVerif's process calculus (which is based on the applied
   pi calculus) into Horn clauses.

   This translator uses some parts of ProVerif, in particular modified versions
   of the parser and some display functions. Other than that, it is a
   stand-alone program that outputs a ProVerif-compatible description of
   the produced Horn clauses.

   For more information about StatVerif, and to obtain a copy of the
   paper, please visit: http://markryan.eu/research/projects/StatVerif/

   Copyright (C) Joshua Phillips, 2011

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details (in file LICENSE).

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)



open Piparser   (* Use the modified pi-calculus parser *)
open Parsing_helper
open Types      (* Types for representing terms, processes etc. *)
open Stringmap  (* A string-keyed associative array *)
open Bindermap  (* A binder-keyed associative array *)

exception No_Process of string ;;
exception Multiple_Processes of string ;;
exception Fact_Not_Equal ;;
exception Process_Not_In_Standard_Form of string ;;
exception Duplicate_Function of string ;;
exception Cell_Not_At_Start ;;
exception Not_Implemented of string ;;
exception Conflicting_Substitutions ;;
exception Undefined_Subprocess of string ;;
exception Undefined_Function of string ;;
exception Undefined_Name of string ;;
exception Bad_Process of string ;;
exception Wrong_Arity of string ;;
exception Unify of term * term ;;
exception Internal_Error of string ;;
exception Not_A_Var of string ;;

(* note: Types.Tuple is used in various places to fill in funcats
   fields, because I don't know what funcats is (are?) supposed
   to represent! *)

(* To replace BinderMap.filter *)
let bindermap_filter f m =
    BinderMap.fold (fun k v m ->
        if (f k v) then
            BinderMap.add k v m
        else
            m
      ) m BinderMap.empty

let bindermap_merge f a b =
    let new_map = ref BinderMap.empty in
    BinderMap.iter (fun k va ->
        let vb =
            if BinderMap.mem k b then
                Some (BinderMap.find k b)
            else
                None
        in
        match (f k (Some va) vb) with
            Some vc ->
                new_map := BinderMap.add k vc !new_map
          | None ->
                ()
      ) a;
    BinderMap.iter (fun k vb ->
        if not (BinderMap.mem k a) then
            match (f k None (Some vb)) with
                Some vc ->
                    new_map := BinderMap.add k vc !new_map
              | None ->
                    ()
      ) b;
    !new_map

let stringmap_bindings m =
    StringMap.fold (fun k v l -> (k,v) :: l) m []

(* Make a new binder (variable name) named s. *)
let string_to_binder s =
    { sname = s;
      vname = 0; (* A number appended to the variable name *)
      btype = { tname = "" }; (* types not implemented *)
      link = NoLink } ;;

(* Return a string representation of a binder. *)
let namestr b =
    if b.vname > 0 then
        b.sname ^ "_" ^ string_of_int b.vname
    else
        b.sname ;;

(* Represents a function declaration. The only things that are
   specified in the input file are the name and arity. *)
type function_declaration = {
    fd_name : string;
    fd_arity : int;
    fd_is_private : bool } ;;

(* global_state represents the last known states of memory cells.
   This is the type used to represent phi, phi0 etc. *)
type global_state = term list ;;

(* A hypothesis is a predicate applied to some terms.
   Message (s, c, x) means that x has been output on channel c in state
   s, while Attacker (s, x) means that an attacker can know x in state
   s. *)
type hypothesis =
    Message of global_state * term * term
  | Attacker of global_state * term ;;

type horn_clause = hypothesis list * hypothesis ;;

type output_item =
    OutputClause of horn_clause * string
  | OutputComment of string ;;

(* A memory cell. This is defined in the process as
   [name |-> initvalue]. *)
type cell = {
    c_name : binder;
    c_init_value : term } ;;

(* A reduction is a tuple (lhs, rhs, is_private). lhs is a destructor applied to a
   list of terms, and rhs is the result of that application. *)
type reduc = term * term * bool ;;

(* A free name, declared in a "free" clause. *)
type free_name = {
    fn_name : string;
    fn_private : bool } ;;

type query_var =
    Unambiguous of term
  | Ambiguous ;;

(* State of the translator. This includes information about declarations
   from the input file, as well as the variables rho, H, l, phi and mu
   which are described in the paper. *)
type translator_state = {
    s_functions : function_declaration StringMap.t;
    s_reductions : reduc list;
    s_subprocesses : Piptree.process StringMap.t;  (* named subprocesses *)
    s_free_names : free_name list; (* free names of the process that are declared in the source file *)
    s_cells : cell list;
    mutable s_varcounter : int;    (* is incremented to produce unique variable names *)
    s_query_vars : query_var StringMap.t ref; (* This is populated with variable names that may be used in queries *)
    s_orig_names : string StringMap.t ref; (* Original names for renamed restrictions *)
    (* s_vars are rho H l phi and mu from the paper. *)
    s_vars : term BinderMap.t (* rho; variable bindings *)
           * hypothesis list        (* H; accumulation of hypotheses *)
           * term list            (* l; variables that have been input *)
           * global_state           (* phi *)
           (*  * bool ;    *)             (* mu; locked or not *)
           * bool list ;
    s_process : process option ;
    s_params: Piptree.decl list } ;;

(* A value for translator_state that represents the starting state for
   translation. *)
let empty_state = {
    s_functions = StringMap.empty ;
    s_reductions = [] ;
    s_subprocesses = StringMap.empty ;
    s_free_names = [] ;
    s_cells = [] ;
    s_varcounter = 1 ;
    s_query_vars = ref StringMap.empty ;
    s_orig_names = ref StringMap.empty ;
    s_vars = (BinderMap.empty, [], [], [], []) ;
    s_process = None ;
    s_params = [] } ;;

(* Create a new function_declaration named 'name' with the given arity,
   add it to the translator_state ts and return the updated ts. *)
let new_function ts name arity is_private =
    (if StringMap.mem name ts.s_functions then
        raise (Duplicate_Function name));
    { ts with s_functions =
        StringMap.add name
            { fd_name = name ;
              fd_arity = arity ;
              fd_is_private = is_private }
            ts.s_functions } ;;

(* Add an item to s_query_vars *)
let add_query_var ts uname value =
    let name =
        try
            StringMap.find uname !(ts.s_orig_names)
        with Not_found ->
            uname in
    if StringMap.mem name !(ts.s_query_vars) then
        ts.s_query_vars := StringMap.add name Ambiguous !(ts.s_query_vars)
    else
        ts.s_query_vars := StringMap.add name (Unambiguous value) !(ts.s_query_vars) ;;

(* Add a new reduction to ts, and return the updated ts. *)
let new_reduction ts reduc =
    { ts with s_reductions = reduc :: ts.s_reductions } ;;

(* Add the declaration of a free name 'name' to ts, and return
   the updated ts. *)
let new_free_name ts freename =
    match ts.s_vars with (rho, h, l, phi, mu) ->
        let b = string_to_binder freename.fn_name in
        { ts with
            s_free_names = freename :: ts.s_free_names;
            s_vars = (
                BinderMap.add b (Bracketed(b, [])) rho,
                h, l, phi, mu) } ;;

(* [a; b; c; ...; z] -> [(0,a); (1,b); (2,c); ...; (n-1,z)] *)
let enumerate lst =
    let counter = ref (-1) in
    List.map (fun x -> counter := !counter + 1; (!counter, x)) lst ;;

(* Return lst with nth element removed. *)
let exclude_nth lst n =
    let rec inner i = (
        if i = (List.length lst) then []
        else if i = n then inner (i + 1)
        else (List.nth lst i) :: (inner (i + 1))) in
    inner 0

(* n -> [0; 1; 2; ...; n-1] *)
let range n =
    let rec construct l = function
        0 -> l
      | n -> construct ((n-1) :: l) (n-1)
    in
        construct [] n ;;

(* used by rep *)
let rec rep_inner n x l =
    match n with
        0 -> l
      | n -> rep_inner (n-1) x (x::l) ;;

(* Return a list of x repeated n times. *)
let rep n x = rep_inner n x [] ;;

(* Return the relative complement of a in b *)
let relative_complement a b =
    let result = ref b in
    List.iter (fun a_item ->
        result := List.filter (fun b_item -> a_item <> b_item) !result)
        a;
    !result ;;

(* Return a funsymb with the given identifier. *)

let string_to_funsymb s = {
    f_name = s;
    f_type = ([], {tname=""}); (* types not implemented *)
    f_cat = Tuple; (* What is this? *)
    f_initial_cat = Tuple;
    f_private = false;
    f_options = 0 } ;; (* What is this? *)

let create_funsymb ident = string_to_funsymb (fst ident) ;;

let is_constructor state func = StringMap.mem func.f_name state.s_functions ;;

(* Functions for applying substitutions to various things.
   A substitution is represented as a term BinderMap. *)

let rec apply_subst_to_term subst term =
    match term with
        Var b ->
            if BinderMap.mem b subst then
                BinderMap.find b subst
            else
                term
      | FunApp(f, terms) -> FunApp(f, apply_subst_to_terms subst terms)
      | x -> x

and apply_subst_to_terms subst terms =
    List.map (apply_subst_to_term subst) terms

(* Apply a substitution to the range of another substitution.
   This is not the same as composition, because in apply_subst_to_subst,
   the domain of apply_to is unchanged. *)
and apply_subst_to_subst mapping apply_to =
    BinderMap.map (apply_subst_to_term mapping) apply_to

(* Substitution composition
   (From page 7 of "Unification Theory",
   Franz Baader and Wayne Snyder).  *)
and compose_substs sigma theta =
    (* Apply theta to every term in Ran(sigma) to obtain sigma1 *)
    let sigma1 = apply_subst_to_subst theta sigma in
    (* Remove from theta any binding x -> t, where x in Dom(sigma), to obtain theta1 *)
    let theta1 = bindermap_filter (fun k _ -> not (BinderMap.mem k sigma)) theta in
    (* Remove from sigma1 any trivial binding x -> x to obtain sigma2 *)
    let sigma2 = bindermap_filter (fun k v ->
        match v with
            Var x when x = k -> false
          | _ -> true) sigma1 in
    (* Take the union of the two sets of bindings sigma2 and theta1 *)
    bindermap_merge (fun k v1 v2 ->
        match v1, v2 with
            x, None -> x
          | None, x -> x
          | x, y when x = y -> x
          | Some x, Some y ->
                print_string ("Conflicting bindings for " ^ (namestr k) ^ ": ");
                Display.display_term x;
                print_string " and ";
                Display.display_term y;
                print_endline ".";
                raise Conflicting_Substitutions)
        sigma2 theta1

(* ... and to hypotheses *)
and apply_subst_to_hyp mapping = function
    Message(phi, c, m) ->
        Message(
            apply_subst_to_terms mapping phi,
            apply_subst_to_term mapping c,
            apply_subst_to_term mapping m)
  | Attacker(phi, m) ->
        Attacker(
            apply_subst_to_terms mapping phi,
            apply_subst_to_term mapping m)

and apply_subst_to_hyps mapping =
    List.map (apply_subst_to_hyp mapping)

and apply_subst_to_clause mapping (hypotheses, conclusion) =
    apply_subst_to_hyps mapping hypotheses, apply_subst_to_hyp mapping conclusion

and apply_subst_to_binder mapping binder =
    if BinderMap.mem binder mapping then
        match BinderMap.find binder mapping with
            Var x -> x
          | _ -> raise (Not_A_Var "Cannot map a binder to a non-binder term.")
    else
        binder ;;
    
(* Print a mapping (substitution), for debugging purposes. *)
let print_mapping m =
    BinderMap.iter (fun k v ->
        print_string ((namestr k) ^ " -> ");
        Display.display_term v;
        print_endline "") m ;;

(* Return true if function f exists,
   regardless of whether it's a constructor or a destructor. *)
let function_exists ts f =
    StringMap.mem f ts.s_functions
    || List.exists (function FunApp(fr, argsr), _, _
        when fr.f_name = f -> true
      | _ -> false) ts.s_reductions ;;

(* Convert an identifier to a binder. *)
let convert_ident ident = string_to_binder (fst ident) ;;

(* Convert a term parse tree (from the parser)
   to a term that we can use. *)
let rec convert_term ts = function
    Piptree.PIdent ident -> Var (convert_ident ident)
  | Piptree.PFunApp (ident, terms) ->
        (if not (function_exists ts (fst ident)) then
            raise (Undefined_Function (fst ident)));
        (if StringMap.mem (fst ident) ts.s_functions then
            let fundecl = StringMap.find (fst ident) ts.s_functions in
            if fundecl.fd_arity <> (List.length terms) then
                raise (Wrong_Arity (fst ident)));
        FunApp (
            { f_name = fst ident;
              f_type = ([], {tname=""}); (* types not implemented *)
              f_cat = Tuple; (* What is this? *)
              f_initial_cat = Tuple;
              f_private = false;
              f_options = 0 }, (* What is this? *)
            (List.map (fun x -> convert_term ts (fst x)) terms))
  | Piptree.PTuple _ -> raise (Not_Implemented "tuples")
    ;;

(* Same as convert_term, but with a couple of differences. It is used
   when processing reduction definitions. The differences are:
    - If a name x is declared as a free name, replace it with x[] so
      that it can't be substituted during unification.
    - If func_is_new is true, an undefined function is not considered
      erroneous. This is because destructors are defined in the
      reductions themselves, and not declared beforehand. For example:
          projl(pair(a,b)) = a;
      may be the first time "projl" appears. *)
let rec convert_term_free ts func_is_new = function
    Piptree.PIdent ident ->
        if List.exists (fun x -> (fst ident) = x.fn_name) ts.s_free_names then
            Bracketed (convert_ident ident, [])
        else
            Var (convert_ident ident)
  | Piptree.PFunApp (ident, terms) ->
        (if not func_is_new then
            (if not (function_exists ts (fst ident)) then
                raise (Undefined_Function (fst ident)));
            (if StringMap.mem (fst ident) ts.s_functions then
                let fundecl = StringMap.find (fst ident) ts.s_functions in
                if fundecl.fd_arity <> (List.length terms) then
                    raise (Wrong_Arity (fst ident))));
        FunApp (
            { f_name = fst ident;
              f_type = ([], {tname=""}); (* types not implemented *)
              f_cat = Tuple; (* What is this? *)
              f_initial_cat = Tuple;
              f_private = false;
              f_options = 0 }, (* What is this? *)
              (* set func_is_new to false here, because only the
                 destructor being defined should be allowed to be
                 undefined. *)
            (List.map (fun x -> convert_term_free ts false (fst x)) terms))
  | Piptree.PTuple _ -> raise (Not_Implemented "tuples")
    ;;

(* Convert a Piptree.pattern to a pattern *)
let rec convert_pattern ts = function
    Piptree.PPatVar(ident) -> PatVar(convert_ident ident)
  | Piptree.PPatTuple _ -> raise (Not_Implemented "tuples")
  | Piptree.PPatFunApp(ident, patterns) -> PatTuple(
        create_funsymb ident,
        List.map (convert_pattern ts) patterns)
 | Piptree.PPatEqual(term) -> PatEqual(convert_term ts (fst term))
    ;;

(* Get the left term of a Piptree.fact as a term *)
let get_fact_left ts = function
    Piptree.PSEqual(a,b) -> convert_term ts (fst a)
  | _ -> raise Fact_Not_Equal ;;

(* Get the right term of a Piptree.fact as a term *)
let get_fact_right ts = function
    Piptree.PSEqual(a,b) -> convert_term ts (fst b)
  | _ -> raise Fact_Not_Equal ;;

(* Get occurrence (line number) of a term or fact *)

let get_occurrence_f fact =
    (fst (snd fact)).Lexing.pos_lnum ;;

let get_occurrence_e e =
    (fst (snd e)).Lexing.pos_lnum ;;

(* Return a list of all the names that occur in the given term. *)
let rec all_names = function
    Var x -> [x]
  | FunApp (_, args) ->
        List.flatten (List.map all_names args)
  | Bracketed (b, []) -> [b]
  | Bracketed (b, l) ->
        (* all_names applied to terms which contain a Bracketed. *)
        raise (Internal_Error "trying to retrieve all_names from a Bracketed") ;;

(* Return true if there are no destructors in the given term. *)
let rec no_destructors ts = function
    FunApp(g, args) ->
        (StringMap.mem g.f_name ts.s_functions)
        && (List.for_all (no_destructors ts) args)
  | _ -> true ;;

(* Raise an exception if the given term contains any destructors *)
let rec check_term state = function
    FunApp(f, args) ->
        if not (is_constructor state f) then
            raise (Bad_Process ("destructor `" ^ f.f_name ^ "' not in a `let';"
                ^ " enable extensions with -e"))
        else
            List.iter (check_term state) args
  | _ -> () ;;

(* Convert a parse tree to a process.
   This function handles restrictions and cell names
   and updates the translation state accordingly.
   Returns (the updated state, the converted process so far).
   is_start is true for the first part of the process that
   conforms to the form specified in IV.A of the paper. This is to
   collect information about memory cells and the restrictions they are
   under.

   NOTE: This is not the translation. The translation state
   ts is the initial translation state being formed. *)

let rec convert_process ts is_start restrictions p =
    match p with
        Piptree.PNil ->
            ts, Nil
      | Piptree.PPar(Piptree.PCell _ as p, q) ->
            let ts, p' = convert_process ts true restrictions p in
            let ts, q' = convert_process ts true restrictions q in
            ts, Par(p', q')
      | Piptree.PPar(p, q) ->
            let ts, p' = convert_process ts false restrictions p in
            let ts, q' = convert_process ts false restrictions q in
            ts, Par(p', q')
      | Piptree.PRepl(p) ->
            (* We cannot handle cells under replication.
               Therefore, beyond here, is_start is false. *)
            let ts, p' = convert_process ts false restrictions p in
            ts, Repl(p', 0)
      | Piptree.PRestr(ident, p) ->
            (* Create a unique name for this restriction.
               TODO: Make sure the produced name really is unique!
               Restrictions get translated to ground terms, e.g. "new a; P"
               may bind a to a[] in P. Since ground terms are global, these
               names need to be uniquified, so that the two "a"s in
               "new a; P" and "new a; Q" don't get confused
               (never get unified). *)

            let orig_name = fst ident in
            let new_name = orig_name ^ "'" ^ (string_of_int ts.s_varcounter) in
            ts.s_varcounter <- ts.s_varcounter + 1;
            ts.s_orig_names := StringMap.add new_name orig_name !(ts.s_orig_names);
            let restrictions' = BinderMap.add (string_to_binder orig_name)
                (Var (string_to_binder new_name)) restrictions in
            let ts, p' = convert_process ts is_start restrictions' p in
            let ts = (if is_start then
                (* add mapping to rho0 *)
                (add_query_var ts new_name (Bracketed(string_to_binder new_name, []));
                match ts.s_vars with (rho, h, l, phi, mu) ->
                {ts with
                    s_vars = (
                        BinderMap.add (string_to_binder new_name)
                            (Bracketed(string_to_binder new_name, [])) rho,
                        h, l, phi, mu)})
                else ts) in
            ts, Restr(string_to_funsymb new_name, p', 0)
      | Piptree.PTest(fact, p, q) ->
            let lhs = get_fact_left ts (fst fact) in
            let rhs = get_fact_right ts (fst fact) in
            let ts, p' = convert_process ts false restrictions p in
            let ts, q' = convert_process ts false restrictions q in
            ts, Test(apply_subst_to_term restrictions lhs,
                     apply_subst_to_term restrictions rhs,
                     p', q', get_occurrence_f fact)
      | Piptree.PInput(channel, pat, p) ->
            let pat' = convert_pattern ts pat in
            let restrictions' = (
                match pat' with
                    PatVar x -> BinderMap.remove x restrictions
                  | _ -> restrictions) in
            let ts, p' = convert_process ts false restrictions' p in
            ts, Input(
                apply_subst_to_term restrictions
                    (convert_term ts (fst channel)),
                pat',
                p', get_occurrence_e channel)
      | Piptree.POutput(channel, value, p) ->
            let ts, p' = convert_process ts false restrictions p in
            ts, Output(
                apply_subst_to_term restrictions
                    (convert_term ts (fst channel)),
                apply_subst_to_term restrictions
                    (convert_term ts (fst value)),
                p', get_occurrence_e channel)
      | Piptree.PLet(pat, term, p, q) ->
            let pat' = convert_pattern ts pat in
            let restrictions' = (
                match pat' with
                    PatVar x -> BinderMap.remove x restrictions
                  | _ -> restrictions) in
            let ts', p' = convert_process ts false restrictions' p in
            let ts', q' = convert_process ts' false restrictions' q in
            ts', Let(
                pat',
                apply_subst_to_term restrictions (convert_term ts (fst term)),
                p', q', get_occurrence_e term)
      | Piptree.PLockAll(p) ->
            let ts, p' = convert_process ts false restrictions p in
            ts, LockAll(p')
      | Piptree.PUnlockAll(p) ->
            let ts, p' = convert_process ts false restrictions p in
            ts, UnlockAll(p')
      | Piptree.PLock(x,p) ->
            let ts, p' = convert_process ts false restrictions p in
            ts, Lock(apply_subst_to_binder restrictions (convert_ident x), p')
      | Piptree.PUnlock(x,p) ->
            let ts, p' = convert_process ts false restrictions p in
            ts, Unlock(apply_subst_to_binder restrictions (convert_ident x), p')
      | Piptree.PReadAs(s, x, p) ->
            let restrictions' = BinderMap.remove (convert_ident x) restrictions in
            let ts, p' = convert_process ts false restrictions' p in
            ts, ReadAs(
                apply_subst_to_binder restrictions (convert_ident s),
                convert_ident x,
                p')
      | Piptree.PCell(s, init) ->
            (match is_start with
                false -> raise Cell_Not_At_Start
              | true ->
                    let sname = apply_subst_to_binder restrictions
                        (convert_ident s) in
                    let sinit = apply_subst_to_term restrictions
                        (convert_term ts (fst init)) in
                    let new_cell = {c_name = sname; c_init_value = sinit} in
                    (match ts.s_vars with rho, h, l, phi, mu ->
                        let ts = { ts with
                            s_cells = (* new_cell :: ts.s_cells *) ts.s_cells @ [new_cell] } in
                        ts, Cell(sname, sinit)))
                    (*let ts3, trans_p = convert_process ts2 true p in
                    ts3, Cell(sname, sinit, trans_p))*)
      | Piptree.PAssign(s, m, p) ->
            let ts, p' = convert_process ts false restrictions p in
            ts, Assign(
                apply_subst_to_binder restrictions (convert_ident s),
                apply_subst_to_term restrictions (convert_term ts (fst m)),
                p',
                get_occurrence_f s)
      | Piptree.PLetDef(ident) ->
            (* Part of the process is defined separately.
               Inline the process here, during conversion. *)
            if StringMap.mem (fst ident) ts.s_subprocesses then
                convert_process ts is_start restrictions
                    (StringMap.find (fst ident) ts.s_subprocesses)
            else
                raise (Undefined_Subprocess (fst ident))
      | _ -> raise (Not_Implemented "some fancy process constructs are not implemented")
;;

(* Process one declaration in the input file. *) 
let initial_state_decl ts = function
    Piptree.FunDecl(ident, arity, is_private) ->
        new_function ts (fst ident) arity is_private
  | Piptree.Reduc(defs, is_private) ->
        List.fold_left
            (fun ts def ->
                (* use convert_term_free here, because the free names must
                   not be substituted. *)
                let lhs = convert_term_free ts true (fst (fst def)) in
                let rhs = convert_term_free ts false (fst (snd def)) in
                (* The lhs should be in the form g(...) with no destructors occurring within the "..." *)
                (match lhs with
                    FunApp(g, args) ->
                        List.iter (fun arg -> if not (no_destructors ts arg) then
                            raise (Bad_Process ("nested destructor in lhs of reduction for `" ^ g.f_name ^ "'")))
                            args
                  | _ -> raise (Internal_Error "lhs of reduction is not a function application"));
                (* The rhs should have no destructors at all. *)
                (if not (no_destructors ts rhs) then
                    raise (Bad_Process ("destructor in rhs of reduction")));
                (* All names in rhs that are not in lhs should be declared free. *)
                let lhs_names = all_names lhs in
                let rhs_names = all_names rhs in
                let freenames = relative_complement lhs_names rhs_names in
                List.iter (fun x ->
                    if not (List.exists (fun s -> s.fn_name = x.sname) ts.s_free_names) then
                        raise (Undefined_Name x.sname))
                    freenames;
                new_reduction ts (lhs, rhs, is_private))
            ts defs
  | Piptree.Free(idents, is_private) ->
        List.fold_left (fun ts ident -> new_free_name ts { fn_name = fst ident; fn_private = is_private } )
            ts idents
  | Piptree.Process(p) ->
        (match ts.s_process with
            None ->
                let ts, process = convert_process ts true BinderMap.empty p in
                assert (ts.s_process = None);
                (match ts.s_vars with (rho0, h, l, junk, mu) ->
                    { ts with
                        s_vars = (
                            rho0, h, l, 
                            (List.map
                                (fun cell -> apply_subst_to_term rho0 cell.c_init_value)
                                ts.s_cells), (* calculate phi0 *)
                            (List.map
                                (fun cell -> false) ts.s_cells));
                        s_process = Some process })
          | _ -> raise (Multiple_Processes "more than one process defined"))
  | Piptree.PDef(ident, process) ->
        {ts with
            s_subprocesses =
                StringMap.add (fst ident) process ts.s_subprocesses}
  | Piptree.Param(k, v) as item ->
        {ts with s_params = item :: ts.s_params}
  | _ -> ts ;;

(* Create the initial translator state from the parse tree. *)
let rec initial_state ts = function
    decl :: rest ->
        initial_state (initial_state_decl ts decl) rest
  | [] -> ts ;;

(* Return true if the name x appears in the given term; false otherwise. *)
let rec var_is_in_term x = function
    Var y -> x = y
  | FunApp(g, args) -> List.exists (var_is_in_term x) args
  | Bracketed(y, args) -> x = y || List.exists (var_is_in_term x) args ;;

let var_is_in_terms x =
    List.exists (var_is_in_term x) ;;

(* Return true if the name x appears in the given hypothesis; false otherwise. *)
let var_is_in_hyp x = function
    Attacker(vs,vm) -> List.exists (var_is_in_term x) vs || var_is_in_term x vm
  | Message(vs,vc,vm) -> List.exists (var_is_in_term x) vs || var_is_in_term x vc || var_is_in_term x vm ;;

(* Return true if the name x appears in any of the given hypotheses; false otherwise *)
let var_is_in_hyps x =
    List.exists (var_is_in_hyp x)

(* Unification of terms *)

(* Unify two terms. Return a substitution. *)
 let rec unify ts = function
    a, b when a = b -> BinderMap.empty
  | Var a, rhs ->
        BinderMap.add a rhs BinderMap.empty
  | lhs, Var a -> unify ts (Var a, lhs)
  | FunApp(f1, args1), FunApp(f2, args2) ->
        if f1 = f2 then
            let subst_so_far = ref BinderMap.empty in
            List.iter2
                (fun a b ->
                    subst_so_far := compose_substs
                        !subst_so_far
                        (unify ts
                            (apply_subst_to_term !subst_so_far a,
                             apply_subst_to_term !subst_so_far b)))
                args1 args2;
            (*List.fold_left merge_substs BinderMap.empty
                (List.map2 (fun a b -> unify ts (a, b)) args1 args2)*)
            !subst_so_far
        else
            raise (Unify (FunApp(f1, args1), FunApp(f2, args2)))
  | lhs, rhs ->
        raise (Unify(lhs,rhs))

;;

(* Find the inner process, i.e. the process P in:
   new mtilde; ([s1->M1] | ... | [sn->Mn] | P) *)
let rec find_inner_2 = function
    Par(p,q) ->
        (match find_inner_2 p, find_inner_2 q with
            Nil, x -> x
          | x, Nil -> x
          | x, y -> Par(p,q))
  | Cell(_,_) -> Nil
  | x -> x ;;

let rec find_inner = function
    Restr(_,p,_) -> find_inner p
  | x -> find_inner_2 x ;;

(* Make a variable name unique by appending a number.
   This is used to make fresh names during translation. *)
let unique_var x n =
    {x with vname = n} ;;

(* Return a list of the bound names of process p *)
let rec bound_names = function
    Nil -> []
  | Output(_,_,p,_) -> bound_names p
  | Input(m,PatVar(x),p,_) -> x :: bound_names p
  | Par(p,q) -> (bound_names p) @ (bound_names q)
  | Restr(fs,p,_) -> (string_to_binder fs.f_name) :: (bound_names p)
  | Repl(p,_) -> bound_names p
  | Test(_,_,p,q,_) -> (bound_names p) @ (bound_names q)
  | Let(PatVar lhs,rhs,p,q,_) -> lhs :: (bound_names p) @ (bound_names q) (* XXX: Is this true? *)
  | LockAll(p) | UnlockAll(p)
  | Lock(_,p) | Unlock(_,p) -> bound_names p
  | ReadAs(s,x,p) -> x :: (bound_names p) (* XXX: Is this true? *)
  | Cell _ -> []
  | Assign(_,_,p,_) -> bound_names p
  | _ -> raise (Internal_Error "unimplemented type of process") ;;

(* Filter out x from lst. *)
let exclude x lst =
    List.filter (fun item -> item <> x) lst ;;

(* Return a list of the free names of process p *)
let rec free_names = function
    Nil -> []
  | Output(a, x, p, _) -> (all_names a) @ (all_names x) @ free_names p
  | Input(a, PatVar x, p, _) -> (all_names a) @ exclude x (free_names p)
  | Par(p,q) -> (free_names p) @ (free_names q)
  | Restr(fs,p,_) -> exclude (string_to_binder fs.f_name) (free_names p)
  | Repl(p,_) -> free_names p
  | Test(_,_,p,q,_) -> (free_names p) @ (free_names q)
  | Let(PatVar lhs,rhs,p,q,_) -> exclude lhs ((free_names p) @ (free_names q)) (* XXX: Is this true? *)
  | LockAll(p) | UnlockAll(p) | Lock(_,p) | Unlock(_,p) -> free_names p
  | ReadAs(s,x,p) -> exclude x (free_names p) (* XXX: Is this true? *)
  | Cell(s,m) -> [s] (* XXX: no idea *)
  | Assign(s, x, p, _) -> (all_names x) @ (free_names p)
  | _ -> raise (Internal_Error "unimplemented type of process") ;;

(* Return all the reductions for destructor g with n args. *)
let find_reductions state g n =
    List.filter (function
        FunApp(fr, argsr), _, _
        when (List.length argsr) = n
        && fr.f_name = g.f_name -> true
      | _ -> false) state.s_reductions ;;

(* Make names in term t unique.
   Returns t', which is t with unique names.
   Merges u into 'mapping', where tu = t'. *)
let rec uniquify_names state mapping t =
    match t with
        Var x ->
            if BinderMap.mem x !mapping then
                (* x is already in mapping - i.e. we already decided what to
                   name this variable. Re-use that name. *)
                BinderMap.find x !mapping
            else
                (* Append a number to the name to make it unique.
                   XXX: This assumes the user hasn't appended their own numbers! *)
                let new_x = { x with vname = state.s_varcounter } in
                state.s_varcounter <- state.s_varcounter + 1;
                print_string "Uniquifying ";
                Display.display_term (Var x);
                print_string " to ";
                Display.display_term (Var new_x);
                print_endline ".";
                mapping := BinderMap.add x (Var new_x) !mapping;
                Var new_x
      | FunApp(f, args) ->
            FunApp(f, List.map (uniquify_names state mapping) args)

      | Bracketed(b, l) ->
            (* This must have already been substituted. *)
            t ;;

(* Return the state terms of a hypothesis *)
let states = function
    Attacker(s, _) -> s
  | Message(s, _, _) -> s ;;

(* Return the jth state term of a hypothesis *)
let stat j hyp = List.nth (states hyp) j ;;

(* Return the non-state terms of a hypothesis. *)
let non_states = function
    Attacker(_, vm) -> [vm]
  | Message(_, vc, vm) -> [vc; vm] ;;

let completely_fresh_phi state =
    List.map (fun cell -> Var(unique_var cell.c_name state.s_varcounter)) state.s_cells ;;

(* For unlocked cells, create fresh variables in phi.
   Returns the updated state. *)
let partly_fresh_phi state =
    match state.s_vars with (rho, h, l, phi, mu) ->
    let new_vars = ref [] in
    let phi' = List.map2 (
        fun phi_item (cell, locked) ->
            if locked then
                phi_item
            else
                (* This cell is unlocked. Create a fresh variable to represent
                   its unknown state. *)
                let new_var = unique_var cell.c_name state.s_varcounter in
                new_vars := new_var :: !new_vars;
                Var(unique_var cell.c_name state.s_varcounter))
        phi (List.combine state.s_cells mu) in
    (* Add the fresh variables to rho. *)
    let rho' = List.fold_left
        (fun m x -> BinderMap.add x (Var x) m)
        rho !new_vars in
    {state with s_varcounter = state.s_varcounter + 1;
                s_vars = (rho', h, l, phi', mu)} ;;

(* Return index (0..length state.s_cells-1)
   of the cell with the name "cell". *)
let cell_index state cell =
    match cell with
        Bracketed(var, []) ->
            let counter = ref (-1) in
            let index = List.fold_left (fun index cell ->
                counter := !counter + 1;
                if cell.c_name = var then !counter
                else index) (-1) state.s_cells in
            (if index = -1 then
                raise (Bad_Process ((namestr var) ^ " is not a cell name")));
            index
      | x ->
            print_string "Not a cell name: ";
            Display.display_term x;
            print_endline ".";
            raise (Bad_Process "Bad cell name.") ;;

(* Transform a global_state by setting the entry corresponding
   to cell name "cell", to the new value "expr". *)
let update_phi state cell expr phi =
    match cell with
        Bracketed(var, []) ->
            let found = ref false in
            let result = List.map2 (fun phi_value cell ->
                if cell.c_name = var then
                    (* Assign to this cell *)
                    (found := true;
                    expr)
                else
                    (* Retain current value *)
                    phi_value) phi state.s_cells in
            (if not !found then
                raise (Bad_Process ((namestr var) ^ " is not a cell name")));
            result
      | x ->
            print_string "Not a cell name: ";
            Display.display_term x;
            print_endline ".";
            raise (Bad_Process "Bad cell name.") ;;

(* Main translation function. This takes the same structure as fig. 3
   of the paper. It just doesn't look quite as neat!

   The initial "state" parameter should be computed by initial_state.
   The process "process" should be found by find_inner. *)

let rec translate state process =
    match state.s_vars, process with
        (rho, h, l, phi, mu), Nil ->
            List.iter (fun (n, mu) ->
                if mu then
                    raise (Bad_Process ("cell "
                        ^ (namestr (List.nth state.s_cells n).c_name)
                        ^ " is locked but never unlocked")))
                (enumerate mu);
            []
      | (rho, h, l, phi, mu), Par(p,q) ->
            (if List.exists (fun x -> x) mu then raise (Bad_Process "parallel inside a lock"));
            translate state p @ translate state q
      | (rho, h, l, phi, mu), Repl(p,_) ->
            (if List.exists (fun x -> x) mu then raise (Bad_Process "replication inside a lock"));
            translate state p
      | (rho, h, l, phi, mu), Input(m,PatVar(x),p,_) ->
            check_term state m;
            let state2 = partly_fresh_phi state in
            (match state2.s_vars with (rho', h', l', phi', mu') ->
            (* Add message hypothesis *)
            let h'' = Message(phi', apply_subst_to_term rho' m, Var x) :: h' in
            let rho'' = BinderMap.add x (Var x) rho' in
            translate { state2 with s_vars = ( rho'', h'', (Var x) :: l',
                phi (* XXX: not phi'?? *) , mu ) } p)
                (* (Using phi' there will probably make no difference anyway) *)
      | (rho, h, l, phi, mu), Output(m,n,p,occurrence) ->
            check_term state m;
            check_term state n;
            (* Try to choose a set of state values that ProVerif will be happy with.
               I think any state values from the hypotheses, or phi, are valid. *)
            OutputClause((List.rev h, Message(phi, apply_subst_to_term rho m, apply_subst_to_term rho n)),
                "output on l." ^ (string_of_int occurrence))
            :: translate state p
      | (rho, h, l, phi, mu), LockAll(p) ->
            (if List.exists (fun x -> x) mu then raise (Bad_Process "lock within a lock"));
            let phi0 = List.map (fun cell -> Var(unique_var cell.c_name state.s_varcounter)) state.s_cells in
            translate { state with
                s_varcounter = state.s_varcounter + 1;
                s_vars = (
                    List.fold_left
                        (fun m x ->
                            match x with Var x -> BinderMap.add x (Var x) m
                            | _ -> raise (Internal_Error "non-variable in phi0"))
                        rho phi0,
                    h, l, phi0, (List.map (fun x -> true) state.s_cells)) }
                p
      | (rho, h, l, phi, mu), UnlockAll(p) ->
            (List.iter2 (fun x cell ->
                if not x then
                    raise (Bad_Process ("cannot unlock all cells: "
                        ^ (namestr cell.c_name) ^ " is not locked")))
                mu state.s_cells);
            translate {state with s_vars = (rho, h, l, phi, List.map (fun x -> false) state.s_cells)} p
      | (rho, h, l, phi, mu), Lock(x,p) ->
            raise (Bad_Process ("partial locking not supported"))
      | (rho, h, l, phi, mu), Unlock(x,p) ->
            raise (Bad_Process ("partial locking not supported"))
      | (rho, h, l, phi, mu), Assign(var, expr, p, occurrence) ->
            (* Create two fresh variables.
               XXX: Make sure the names we create aren't already in rho!
               Note: s_varcounter is incremented by partly_fresh_phi *)

            check_term state expr;

            let vc = string_to_binder ("vc" ^ (string_of_int state.s_varcounter)) in
            let vm = string_to_binder ("vm" ^ (string_of_int state.s_varcounter)) in
            let state2 = partly_fresh_phi state in
            (match state2.s_vars with (rho', h', l', phi0, mu') ->
            let phi1 = update_phi state (apply_subst_to_term rho' (Var var)) (apply_subst_to_term rho' expr) phi0 in
            (* For unlocked cells, translate P with the original phi entry.
               For locked cells, use the new entry. *)
            let next_phi = List.map2 (fun (phi_var, phi1_var) locked -> if locked then phi1_var else phi_var)
                (List.combine phi phi1) mu in
            OutputClause((List.rev (Message(phi0, Var vc, Var vm) :: h), Message(phi1, Var vc, Var vm)),
                "assignment on l." ^ (string_of_int occurrence))
            :: OutputClause((List.rev (Attacker(phi0, Var vm) :: h), Attacker(phi1, Var vm)),
                "assignment on l." ^ (string_of_int occurrence))
            :: (translate
                {state2 with
                    s_vars = (
                        BinderMap.add vc (Var vc)
                            (BinderMap.add vm (Var vm) rho'),
                        h', l', next_phi, mu)}
                p))
      | (rho, h, l, phi, mu), ReadAs(statevar,dest,p) ->
            (* Create two fresh variables.
               XXX: Make sure the names we create aren't already in rho!
               This would only happen if the input file contains
               variable names like "vc1", "vm1" etc. *)
            let vc = string_to_binder ("vc" ^ (string_of_int state.s_varcounter)) in
            let vm = string_to_binder ("vm" ^ (string_of_int state.s_varcounter)) in
            let state2 = partly_fresh_phi state in
            (match state2.s_vars with (rho', h', l', phi0, mu') ->
            translate
                {state2 with
                    s_vars = (
                        (* obtain relevant value from phi0 *)
                        BinderMap.add dest (List.nth phi0 (cell_index state2 (apply_subst_to_term rho (Var statevar))))
                            (BinderMap.add vc (Var vc)
                                (BinderMap.add vm (Var vm) rho')),
                        (Message(phi0, Var vc, Var vm)) :: h', l', phi0, mu')}
                p)
      | (rho, h, l, phi, mu), Test(m, n, p, q, _) ->
            check_term state m;
            check_term state n;
            print_endline "Translating an 'if':";
            Display.display_process "    " process;
            let lhs = apply_subst_to_term rho m in
            let rhs = apply_subst_to_term rho n in
            print_string "After applying rho: if ";
            Display.display_term lhs;
            print_string " = ";
            Display.display_term rhs;
            print_endline " then";
            let ok, sigma = (
                try
                    true, unify state (lhs, rhs)
                with Unify _ ->
                    false, BinderMap.empty) in
            if not ok then []
            else (
                print_endline "sigma =";
                print_mapping sigma;
                (* translate if part *)
                (translate
                    {state with
                        s_vars = (
                            apply_subst_to_subst sigma rho,
                            apply_subst_to_hyps sigma h,
                            apply_subst_to_terms sigma l,
                            apply_subst_to_terms sigma phi,
                            mu)}
                    p))
            (* translate else part *)
            @ (translate_ignorenil state q)
      | (rho, h, l, phi, mu), Let(PatVar x, FunApp(g, args), p, q, _) ->
            (if not (List.for_all (no_destructors state) args) then
                raise (Bad_Process "nested destructor on rhs of let"));
            if is_constructor state g then
                (* g is a constructor *)
                raise (Bad_Process ("rhs of 'let' is a constructor, not a destructor"))
            else
                (print_endline "Translating a 'let':";
                Display.display_process "    " process;
                let rhoed_args = apply_subst_to_terms rho args in
                print_string ("After applying rho: let " ^ (namestr x) ^ " = ");
                Display.display_term (FunApp(g, rhoed_args));
                print_endline " in";
                let reductions = find_reductions state g (List.length args) in
                print_endline ("Found " ^ (string_of_int (List.length reductions)) ^ " reductions for " ^ g.f_name);
                (List.fold_left List.append [] (List.map (function (* for each reduction *)
                         FunApp(_, pees'), reduc_rhs, _ as reduction ->
                            (match reduction with lhs, rhs, is_private ->
                                print_string "Reduction: ";
                                Display.display_term lhs;
                                print_string " -> ";
                                Display.display_term rhs;
                                print_endline ".");
                            (* "pees'" means p'1, p'2, etc. *)
                            (* Rename variables in the reduction to be unique.
                               We can then unify without worrying about variable names being
                               used from the reduction when they're already in use. *)
                            (* If reduc_rhs is a single variable, rename it to x.
                               This makes the output nicer, because names are used from the source
                               process instead of the reductions. e.g. in 'let xl = projl(x) in P',
                               { xl -> xl } will be used instead of { xl -> xleft_123 }. *)
                            let sigma' = ref (
                                match reduc_rhs with
                                    Var y -> BinderMap.add y (Var x) BinderMap.empty
                                  | _ -> BinderMap.empty) in
                            let pees'_unique = List.map (uniquify_names state sigma') pees' in
                            let ok, sigma = (
                                try
                                    true, unify state
                                        (FunApp (g, pees'_unique),
                                          FunApp (g, rhoed_args))
                                with Unify _ -> false, BinderMap.empty) in
                            if not ok then [] (* if unification failed, ignore this reduction *)
                            else (
                                print_endline "sigma' =";
                                print_mapping !sigma';
                                print_endline "sigma =";
                                print_mapping sigma;
                                print_string "reduc_rhs =";
                                Display.display_term (apply_subst_to_term !sigma' reduc_rhs);
                                print_endline "";
                                let reduc_result = apply_subst_to_term sigma (apply_subst_to_term !sigma' reduc_rhs) in
                                check_term state reduc_result;
                                let rho_addition = BinderMap.add x reduc_result in
                                print_endline "Adding the lhs to rho:";
                                print_mapping (rho_addition BinderMap.empty);
                                print_endline "rho is now:";
                                print_mapping (rho_addition (apply_subst_to_subst sigma rho));
                                (translate
                                    {state with
                                        s_vars = (
                                            rho_addition (apply_subst_to_subst sigma rho),
                                            apply_subst_to_hyps sigma h,
                                            apply_subst_to_terms sigma l,
                                            apply_subst_to_terms sigma phi,
                                            mu)}
                                    p))
                        | _ -> raise (Internal_Error "left side of reduction is not a function application")
                    ) reductions))
                    @ translate_ignorenil state q) (* else part *)
      | (rho, h, l, phi, mu), Let(PatVar x, term, p, q, _) ->
            raise (Bad_Process ("rhs of 'let' is not a destructor"))
      | _, Cell _ -> raise (Not_Implemented "Cell not allowed this late in process")
      | (rho, h, l, phi, mu), Restr(fs, p, _) ->
            (* Find bound names of P. P is the process initially given to
               the translate function, as shown on page 5 of the paper.
               Note that the bound names of P won't include the names in
               mtilde. *)
            (match state.s_process with None -> raise (Internal_Error "s_process is None")
            | Some s_process ->
            let bn = bound_names s_process in (* XXX: s_process is not the inner process! *)
            if List.exists (fun x -> fs.f_name = x.sname) bn then
                (add_query_var state fs.f_name (Bracketed(string_to_binder fs.f_name, l));
                translate
                    {state with
                        s_vars = (
                            BinderMap.add (string_to_binder fs.f_name)
                                (Bracketed(string_to_binder fs.f_name, l)) rho,
                            h, l, phi, mu)}
                    p)
            else
                (* XXX: The definition of bound_names would appear to be such that if Q
                   is a subprocess of P, bn(P) is a superset of bn(Q).
                   Therefore, if we're translating part of the process
                   state.s_process, this part of the translation will be unreachable.
                   Perhaps I haven't defined bound_names correctly. *)
                translate
                    {state with
                        s_vars = (
                            BinderMap.add (string_to_binder fs.f_name)
                                (Bracketed(string_to_binder "attn", [])) rho,
                            h, l, phi, mu)}
                    p)

      | _ ->
        print_endline "Unmatched process:";
        Display.display_process "    " process;
        raise (Not_Implemented "unmatched process")
        
and translate_ignorenil state process =
    (* This is used to ignore the lack of "unlock" in "else" clauses. *)
    match process with
        Nil -> []
      | _ -> translate state process
;;
            
(* Like List.iter but prints 's' between iterations
   (but not at the start or end). This is used to print
   delimited lists of things. *)
let iter_with_delim f lst delim =
    let counter = ref 0 in
    List.iter (fun x ->
        (if !counter > 0 then print_string delim);
        counter := !counter + 1;
        f x) lst ;;

let display_hypothesis = function
    Message(phi, c, x) ->
        print_string "mess:";
        iter_with_delim (fun term -> Display.display_term term) (phi @ c :: [x]) ","
  | Attacker(phi, x) ->
        print_string "att:";
        iter_with_delim (fun term -> Display.display_term term) (phi @ [x]) ","

let display_clauses clauses =
    let n = List.length clauses in
    List.iter (
        fun (i, clause) -> match clause with
            OutputClause((hypotheses, concl), comment) ->
                (match hypotheses with
                    head :: tail ->
                        display_hypothesis head;
                        List.iter (fun hyp -> print_string " & "; display_hypothesis hyp) tail;
                        print_string " -> "
                  | [] -> ());
                display_hypothesis concl;
                print_string (if i = n - 1 then "." else ";");
                (if comment <> "" then
                    print_string (" (* " ^ comment ^ " *)"));
                print_endline "";
          | OutputComment(s) ->
                print_endline "";
                print_endline ("(* " ^ s ^ " *)"))
        (enumerate clauses)
;;

(* Return [vs_cell0; vs_cell1; ...; y; ...; vs_celln]
   i.e. the same as cellvars but with one item changed. *)
let rec cellvars_with_subst ts vs i y =
    List.map (fun (i', x) ->
        if i = i' then y
        else x)
        (enumerate (cellvars ts vs))

(* Return [vs_cell0; vs_cell1; ...] for each cell name cell0, cell1, ... *)
and cellvars ts vs =
    List.map (one_cell_var ts vs) (range (List.length ts.s_cells))

and one_cell_var ts vs i =
    Var (string_to_binder (vs ^ "_" ^ (namestr (List.nth ts.s_cells i).c_name))) ;;

let output_params ts =
    List.iter (function
        Piptree.Param(k, Ptree.S v) ->
            print_endline ("param " ^ (fst k) ^ " = " ^ (fst v) ^ ".")
      | Piptree.Param(k, Ptree.I v) ->
            print_endline ("param " ^ (fst k) ^ " = " ^ (string_of_int v) ^ ".")
      ) ts.s_params

let output_initial_decls ts =
    print_endline ("pred att/" ^ (string_of_int((List.length ts.s_cells) + 1)) ^ ".");
    print_endline ("pred mess/" ^ (string_of_int((List.length ts.s_cells) + 2)) ^ ".");
    StringMap.iter (fun _ fd ->
        print_endline ("fun " ^ fd.fd_name ^ "/"
            ^ (string_of_int fd.fd_arity) ^ "."))
        ts.s_functions;
    print_endline "";
;;

(* Return list of queries. *)
let find_queries decls =
    List.fold_left (
        fun result decl ->
            match decl with
              | Piptree.Query l -> l @ result
              | _ -> result)
        [] decls

(* Return list of secrecy assumptions. *)
let find_secrecy_assumptions decls =
    List.fold_left (
        fun result decl ->
            match decl with
              | Piptree.Not(x,y) -> (x,y) :: result
              | _ -> result)
        [] decls

let rec lookup_query_var ts ident =
    match ts.s_vars with (rho, _, _, _, _) ->
    if BinderMap.mem (convert_ident ident) rho then
        BinderMap.find (convert_ident ident) rho
    else if not (StringMap.mem (fst ident) !(ts.s_query_vars)) then
        Var (convert_ident ident)
    else
        match StringMap.find (fst ident) !(ts.s_query_vars) with
            Ambiguous -> raise (Undefined_Name ("name `" ^ (fst ident) ^ "' in query is ambiguous"))
          | Unambiguous x -> x

let rec display_query ts (Piptree.PBefore(event, hypelem)) =
    display_event ts event

and display_event ts (gfact, _) =
    display_gfact ts gfact

and display_gfact ts (Piptree.PGSimpleFact(ident, terms), _) =
    print_string ((fst ident) ^ ":");
    iter_with_delim (display_gterm ts) terms ","

and display_gterm ts (gterm, _) =
    match gterm with
        Piptree.PGIdent(ident) -> Display.display_term (lookup_query_var ts ident)
      | Piptree.PGFunApp(ident, terms) ->
            (* TODO: validate function's existence *)
            print_string ((fst ident) ^ "(");
            iter_with_delim (display_gterm ts) terms ",";
            print_string ")"
      | Piptree.PGName(ident, []) ->
            print_string ((fst ident) ^ "[]")

;;

let output_queries ts decls =
    List.iter
        (fun (Piptree.PRealQuery query) ->
            print_string "query ";
            display_query ts query;
            print_endline ".";
            print_endline "")
        (find_queries decls)

let output_secrecy_assumptions ts decls =
    List.iter
        (fun ((fact, _), _) ->
            print_string "not ";
            display_gfact ts fact;
            print_endline ".";
            print_endline "")
    (find_secrecy_assumptions decls)

(* Output clauses corresponding to reductions,
   free names etc. See section IV.A.2 .. IV.A.3. *)
let initial_clauses ts =
    let x = string_to_binder "x" in
    let y = string_to_binder "y" in
    let xc = string_to_binder "xc" in
    let xm = string_to_binder "xm" in
    let vc = string_to_binder "vc" in
    let vm = string_to_binder "vm" in
    match ts.s_vars with (rho0, _, _, phi0, _) ->
    let rho0phi0 = apply_subst_to_terms rho0 phi0 in

    [OutputComment "Clauses corresponding to mutability of cells"] @
    List.flatten (List.map (fun (i, cell) ->
        OutputClause(
            ([Attacker(cellvars ts "vs", Bracketed(cell.c_name, []))],
             Attacker(cellvars ts "vs", one_cell_var ts "vs" i)),
            "reading " ^ (namestr cell.c_name)) ::
        OutputClause(
            ([Attacker(cellvars ts "vs", Bracketed(cell.c_name, []));
              Attacker(cellvars ts "vs", Var y);
              Message(cellvars ts "vs", Var xc, Var xm)],
             Message(cellvars_with_subst ts "vs" i (Var y), Var xc, Var xm)),
            "writing " ^ (namestr cell.c_name)) ::
        OutputClause(
            ([Attacker(cellvars ts "vs", Bracketed(cell.c_name, []));
              Attacker(cellvars ts "vs", Var y);
              Attacker(cellvars ts "vs", Var xm)],
             Attacker(cellvars_with_subst ts "vs" i (Var y), Var xm)),
             "writing " ^ (namestr cell.c_name)) :: [])
                            (* Omit these clauses for cells with bound names, because:
                               1. Cells can only be read or written by names, not by variables.
                               2. The attacker cannot obtain and then use cell names. *)
        (List.filter (fun (i, cell) -> (List.exists (fun x -> x.fn_name = cell.c_name.sname && not (x.fn_private)) ts.s_free_names)) (enumerate ts.s_cells))) @

    [OutputComment "Clauses corresponding to the attacker's ability to apply constructors"] @
    List.map (fun (_, fundecl) ->
        OutputClause(
            (
                List.map (fun i ->
                    Attacker(cellvars ts "xs", Var(unique_var x (i+1))))
                    (range fundecl.fd_arity),
                Attacker(cellvars ts "xs",
                    FunApp(
                        string_to_funsymb fundecl.fd_name,
                        List.map (fun i -> Var(unique_var x (i+1))) (range fundecl.fd_arity)))),
            "applying " ^ fundecl.fd_name))
        (List.filter (fun (_, fundecl) -> not fundecl.fd_is_private)
            (stringmap_bindings ts.s_functions)) @

    [OutputComment "Clauses corresponding to the attacker's ability to apply reductions"] @
    List.map (
        function
            FunApp(g, args), rhs, _ ->
                OutputClause(
                    (
                        List.map (fun (i, arg) ->
                            Attacker(cellvars ts "xs", arg))
                            (enumerate args),
                        Attacker(cellvars ts "xs", rhs)),
                    "applying " ^ g.f_name)
          | _ -> raise (Internal_Error "wrong type of term in s_reductions"))
        (List.filter (fun (_, _, is_private) -> not is_private) ts.s_reductions) @

    [OutputComment "Clauses corresponding to the attacker's ability to use channels it knows";
     OutputClause(([Message(cellvars ts "vs", Var vc, Var vm); Attacker(cellvars ts "vs", Var vc)], Attacker(cellvars ts "vs", Var vm)), "receiving");
     OutputClause(([Attacker(cellvars ts "vs", Var vc); Attacker(cellvars ts "vs", Var vm)], Message(cellvars ts "vs", Var vc, Var vm)), "sending");
     OutputComment "Knowledge of free names";
     OutputClause(([], Attacker(rho0phi0, Bracketed(string_to_binder "attch", []))), "attacker's own channel");
     OutputClause(([], Attacker(rho0phi0, Bracketed(string_to_binder "attn", []))), "attacker's own nonce")] @

    List.map (fun fn ->
        OutputClause(([], Attacker(rho0phi0, Bracketed(string_to_binder fn.fn_name, []))), ""))
        (List.filter (fun fn -> not fn.fn_private) ts.s_free_names)
;;

(* Check free names in the process to make sure they're declared. *)
let check state process =
    let fn = free_names process in
    List.iter (fun x ->
        if not (List.exists (fun s -> s.fn_name = x.sname) state.s_free_names) then
            raise (Undefined_Name x.sname))
        fn ;;

(* Program entry point. *)
let _ =
    let input_filename = ref None in
    (* Parse command-line arguments. *)
    Arg.parse []
       (fun x -> input_filename := Some x)
       "Usage: stattrans [options...] input-file";

    match !input_filename with
        None ->
            (prerr_endline "No input file specified.";
            prerr_endline "See --help for usage information.";
            exit 1)
      | Some ifname ->
        (* Open the input file and start lexing. *)
        let lexbuf = Lexing.from_channel (open_in ifname) in
        let decls =
            try
                (* Parse the entire file. *)
                Piparser.all Pilexer.token lexbuf
            with Parsing.Parse_error ->
                input_error "Syntax error" (extent lexbuf)
        in
            print_endline "(*";

            (* Calculate the initial translation state. *)
            let initstate = initial_state empty_state decls in
            match initstate.s_process with None -> raise (No_Process "no process defined")
            | Some process ->

            (* Find the process to translate. *)
            let p_inner = find_inner process in
            (match p_inner with
                Nil -> raise (Process_Not_In_Standard_Form "process is not in the standard form, or is 0")
              | _ -> ());

            (* Check free names in the process. *)
            check initstate process;

            (* Display the entire process, for debugging. *)
            Display.display_process "    " p_inner;

            (* Perform translation. *)
            let clauses = translate initstate p_inner in

            let raw_clauses = (initial_clauses initstate)
                @ OutputComment "Clauses from the process" :: clauses in

            let processed_clauses = ref raw_clauses in

            print_endline "*)";
            print_endline "";
            print_endline "(*****  OUTPUT  *****)";
            (* Output the ProVerif clauses to standard output. *)
            output_params initstate;
            output_initial_decls initstate;
            output_queries initstate decls;
            output_secrecy_assumptions initstate decls;
            print_endline "reduc";
            print_endline "";
            display_clauses !processed_clauses
